from django.contrib import admin, messages
from django.utils.translation import gettext_lazy as _
from leaflet.admin import LeafletGeoAdmin

from .models import Drawing, Layer


class LayerInline(admin.TabularInline):
    model = Layer
    fields = ("name", "color_field", "linetype")
    extra = 0

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.exclude(is_block=True)


@admin.register(Drawing)
class DrawingAdmin(LeafletGeoAdmin):
    list_display = (
        "title",
        "epsg",
    )
    inlines = [
        LayerInline,
    ]

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not obj.epsg:
            messages.add_message(
                request,
                messages.WARNING,
                _(
                    """GeoData missing. Upload a DXF with GeoData,
                    a Parent Drawing or select a Reference Point on the map"""
                ),
            )



# Generated by Django 5.1.1 on 2024-11-14 17:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0004_entity"),
    ]

    operations = [
        migrations.AddField(
            model_name="entity",
            name="block",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="related_insertions",
                to="djeocad.layer",
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-18 13:36

import colorfield.fields
import django.db.models.deletion
import djgeojson.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0002_alter_drawing_options_drawing_designx_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Layer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50, verbose_name="Layer name")),
                (
                    "color_field",
                    colorfield.fields.ColorField(
                        default="#FFFFFF", image_field=None, max_length=25, samples=None
                    ),
                ),
                (
                    "linetype",
                    models.BooleanField(
                        default=True, verbose_name="Continuous linetype"
                    ),
                ),
                ("is_block", models.BooleanField(default=False, editable=False)),
                ("geom", djgeojson.fields.GeometryCollectionField(null=True)),
                (
                    "drawing",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="related_layers",
                        to="djeocad.drawing",
                        verbose_name="Drawing",
                    ),
                ),
            ],
            options={
                "verbose_name": "Layer",
                "verbose_name_plural": "Layers",
                "ordering": ("name",),
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-16 13:48

from django.db import migrations, models

import djeocad.models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0007_entity_rotation_entity_xscale_entity_yscale"),
    ]

    operations = [
        migrations.AlterField(
            model_name="entity",
            name="data",
            field=models.JSONField(default=djeocad.models.get_default_entity_data),
        ),
    ]



from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from djeocad.models import Entity, EntityData, Layer


class Command(BaseCommand):
    help = """
        This command has to be executed when passing from version 0.4.0 to 0.5.0
        of django-geocad. The command populates:
        - 'block' field from the data['block'] key
        - 'xscale' field from the data['xscale'] key
        ...
        It also generates EntityData entries with attribute key/values
    """

    def handle(self, *args, **options):
        if settings.DEBUG is False:
            raise CommandError("This command cannot be run when DEBUG is False.")
        self.stdout.write(
            "Moving Data from Entity JSONField to corresponding fields / models"
        )
        populate_fields()

        self.stdout.write("Done.")


def populate_fields():
    for ent in Entity.objects.all():
        if ent.data:
            if "processed" in ent.data:
                continue
            elif "Block" in ent.data:
                try:
                    block = Layer.objects.get(
                        name=ent.data["Block"], drawing=ent.layer.drawing
                    )
                    ent.block = block
                    if "X scale" in ent.data:
                        ent.xscale = ent.data["X scale"]
                    if "Y scale" in ent.data:
                        ent.yscale = ent.data["Y scale"]
                    if "Rotation" in ent.data:
                        ent.rotation = ent.data["Rotation"]
                    ent.data["processed"] = "true"
                    ent.save()
                    if "attributes" in ent.data:
                        for key, value in ent.data["attributes"].items():
                            EntityData.objects.create(
                                entity=ent,
                                key=key,
                                value=value,
                            )
                    continue
                except ObjectDoesNotExist:
                    continue
            for key, value in ent.data.items():
                EntityData.objects.create(
                    entity=ent,
                    key=key,
                    value=value,
                )
        else:
            ent.data = {}
        ent.data["processed"] = "true"
        ent.save()



from math import atan2, cos, degrees, radians, sin

import ezdxf
import nh3
from colorfield.fields import ColorField
from django.conf import settings
from django.core.validators import FileExtensionValidator
from django.db import IntegrityError, models, transaction
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _
from djgeojson.fields import GeometryCollectionField, PointField
from easy_thumbnails.files import get_thumbnailer
from ezdxf.addons import geo
from ezdxf.lldxf.const import InvalidGeoDataException
from PIL import ImageColor
from pyproj import Transformer
from pyproj.aoi import AreaOfInterest
from pyproj.database import query_utm_crs_info
from shapely.geometry import Point, shape
from shapely.geometry.polygon import Polygon


class Drawing(models.Model):

    title = models.CharField(
        _("Name"),
        help_text=_("Name of the drawing"),
        max_length=50,
    )
    dxf = models.FileField(
        _("DXF file"),
        max_length=200,
        upload_to="uploads/djeocad/dxf/",
        validators=[
            FileExtensionValidator(
                allowed_extensions=[
                    "dxf",
                ]
            )
        ],
    )
    parent = models.ForeignKey(
        "self",
        on_delete=models.SET_NULL,
        related_name="parent_drawing",
        verbose_name=_("Parent Drawing"),
        null=True,
        blank=True,
    )
    image = models.ImageField(
        _("Image"), upload_to="uploads/djeocad/images/", null=True, blank=True
    )
    geom = PointField(_("Location"), null=True, blank=True)
    designx = models.FloatField(
        _("Design point X..."),
        default=0,
    )
    designy = models.FloatField(
        _("...Y"),
        default=0,
    )
    rotation = models.FloatField(
        _("Rotation"),
        default=0,
    )
    epsg = models.IntegerField(
        _("CRS code"),
        null=True,
        editable=False,
    )

    class Meta:
        verbose_name = _("Drawing")
        verbose_name_plural = _("Drawings")

    __original_dxf = None
    __original_geom = None
    __original_designx = None
    __original_designy = None
    __original_rotation = None
    # blacklists in settings
    layer_blacklist = settings.CAD_LAYER_BLACKLIST
    name_blacklist = settings.CAD_BLOCK_BLACKLIST
    entity_types = [
        "POINT",
        "LINE",
        "LWPOLYLINE",
        "POLYLINE",
        "3DFACE",
        "CIRCLE",
        "ARC",
        "ELLIPSE",
        "SPLINE",
        "HATCH",
    ]
    # TEXT overrides MTEXT
    text_types = [
        "MTEXT",
        "TEXT",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__original_dxf = self.dxf
        self.__original_geom = self.geom
        self.__original_designx = self.designx
        self.__original_designy = self.designy
        self.__original_rotation = self.rotation

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return reverse("djeocad:drawing_detail", kwargs={"pk": self.id})

    @property
    def popupContent(self):
        url = self.get_absolute_url()
        title_str = f'<a href="{url}"><strong>{self.title}</strong></a>'
        image = self.image
        if not image:
            return {"content": title_str}
        thumbnailer = get_thumbnailer(image)
        thumb = thumbnailer.get_thumbnail({"size": (256, 192), "crop": True})
        image_str = '<img src="%(image)s">' % {"image": thumb.url}
        return {"content": image_str + "<br>" + title_str}

    def save(self, *args, **kwargs):
        # save and eventually upload DXF
        super().save(*args, **kwargs)
        # check if we have coordinate system
        if not self.epsg:
            # check if user has inserted parent
            if self.parent:
                self.get_geodata_from_parent(*args, **kwargs)
                self.extract_dxf(doc=None, refresh=True)
                return
            # check if user has inserted origin on map
            elif self.geom:
                self.get_geodata_from_geom(*args, **kwargs)
                self.extract_dxf(doc=None, refresh=True)
                return
            # no user input, search for geodata in dxf
            else:
                doc = self.get_geodata_from_dxf(*args, **kwargs)
                # if successful use geodata
                if doc:
                    self.extract_dxf(doc)
                return
        # ok, we have coordinate system
        # check if user has inserted new parent
        if self.parent:
            self.delete_all_layers()
            self.get_geodata_from_parent(*args, **kwargs)
            self.extract_dxf(doc=None, refresh=True)
            return
        # check if user has modified origin on map
        if self.geom and self.__original_geom != self.geom:
            self.delete_all_layers()
            self.get_geodata_from_geom(*args, **kwargs)
            self.extract_dxf(doc=None, refresh=True)
            return
        # check if user changed dxf
        if self.__original_dxf != self.dxf:
            self.delete_all_layers()
            doc = self.get_geodata_from_dxf(*args, **kwargs)
            # if successful use new geodata
            if doc:
                self.extract_dxf(doc)
            # else use old geodata
            elif self.geom:
                self.extract_dxf(doc=None, refresh=True)
            return
        # check if something else changed
        if (
            self.__original_designx != self.designx
            or self.__original_designy != self.designy
            or self.__original_rotation != self.rotation
        ):
            self.delete_all_layers()
            self.extract_dxf(doc=None, refresh=True)

    def delete_all_layers(self):
        all_layers = self.related_layers.all()
        if all_layers.exists():
            all_layers.delete()

    def get_geodata_from_parent(self, *args, **kwargs):
        self.geom = self.parent.geom
        self.epsg = self.parent.epsg
        self.designx = self.parent.designx
        self.designy = self.parent.designy
        self.rotation = self.parent.rotation
        self.parent = None
        super().save(*args, **kwargs)

    def get_geodata_from_geom(self, *args, **kwargs):
        utm_crs_list = query_utm_crs_info(
            datum_name="WGS 84",
            area_of_interest=AreaOfInterest(
                west_lon_degree=self.geom["coordinates"][0],
                south_lat_degree=self.geom["coordinates"][1],
                east_lon_degree=self.geom["coordinates"][0],
                north_lat_degree=self.geom["coordinates"][1],
            ),
        )
        self.epsg = utm_crs_list[0].code
        super().save(*args, **kwargs)

    def get_geodata_from_dxf(self, *args, **kwargs):
        doc = ezdxf.readfile(self.dxf.path)
        msp = doc.modelspace()
        geodata = msp.get_geodata()
        if geodata:
            # check if valid XML and axis order
            try:
                self.epsg, axis = geodata.get_crs()
                if not axis:
                    return False
            except InvalidGeoDataException:
                return False
            utm2world = Transformer.from_crs(self.epsg, 4326, always_xy=True)
            world_point = utm2world.transform(
                geodata.dxf.reference_point[0], geodata.dxf.reference_point[1]
            )
            self.geom = {"type": "Point", "coordinates": world_point}
            self.designx = geodata.dxf.design_point[0]
            self.designy = geodata.dxf.design_point[1]
            self.rotation = degrees(
                atan2(
                    geodata.dxf.north_direction[0],
                    geodata.dxf.north_direction[1],
                )
            )
            super().save(*args, **kwargs)
            return doc
        return False

    def extract_dxf(self, doc=None, refresh=False):
        # prepare transformers
        world2utm, utm2world, utm_wcs, rot = self.prepare_transformers()
        # get DXF if none
        if not doc:
            doc = ezdxf.readfile(self.dxf.path)
        msp = doc.modelspace()
        geodata = msp.get_geodata()
        if not geodata or refresh:
            # faking geodata
            geodata = msp.new_geodata()
            geodata = self.fake_geodata(geodata, utm_wcs, rot)
            # replace stored DXF
            doc.saveas(filename=self.dxf.path, encoding="utf-8", fmt="asc")
        # get transform matrix from true or fake geodata
        m, epsg = geodata.get_crs_transformation(no_checks=True)
        layer_table = self.prepare_layer_table(doc)
        for e_type in self.entity_types:
            self.extract_entities(msp, e_type, m, utm2world, layer_table)
        self.create_layer_entities(layer_table)
        block_table = self.save_blocks(doc, m, utm2world)
        # extract insertions
        for ins in msp.query("INSERT"):
            self.extract_insertions(ins, msp, m, utm2world, layer_table, block_table)

    def prepare_transformers(self):
        world2utm = Transformer.from_crs(4326, self.epsg, always_xy=True)
        utm2world = Transformer.from_crs(self.epsg, 4326, always_xy=True)
        utm_wcs = world2utm.transform(
            self.geom["coordinates"][0], self.geom["coordinates"][1]
        )
        rot = radians(self.rotation)
        return world2utm, utm2world, utm_wcs, rot

    def fake_geodata(self, geodata, utm_wcs, rot):
        geodata.coordinate_system_definition = self.get_epsg_xml()
        geodata.dxf.design_point = (self.designx, self.designy, 0)
        geodata.dxf.reference_point = utm_wcs
        geodata.dxf.north_direction = (sin(rot), cos(rot))
        return geodata

    def get_epsg_xml(self):
        xml = """<?xml version="1.0"
encoding="UTF-16" standalone="no" ?>
<Dictionary version="1.0" xmlns="http://www.osgeo.org/mapguide/coordinatesystem">
<Alias id="%(epsg)s" type="CoordinateSystem">
<ObjectId>EPSG=%(epsg)s</ObjectId>
<Namespace>EPSG Code</Namespace>
</Alias>
<Axis uom="METER">
<CoordinateSystemAxis>
<AxisOrder>1</AxisOrder>
<AxisName>Easting</AxisName>
<AxisAbbreviation>E</AxisAbbreviation>
<AxisDirection>east</AxisDirection>
</CoordinateSystemAxis>
<CoordinateSystemAxis>
<AxisOrder>2</AxisOrder>
<AxisName>Northing</AxisName>
<AxisAbbreviation>N</AxisAbbreviation>
<AxisDirection>north</AxisDirection>
</CoordinateSystemAxis>
</Axis>
</Dictionary>""" % {
            "epsg": self.epsg
        }
        return xml

    def prepare_layer_table(self, doc):
        layer_table = {}
        for layer in doc.layers:
            if layer.dxf.name in self.layer_blacklist:
                continue
            if layer.rgb:
                color = cad2hex(layer.rgb)
            else:
                color = cad2hex(layer.color)
            # get or create used to pass the tests
            layer_obj, created = Layer.objects.get_or_create(
                drawing_id=self.id,
                name=layer.dxf.name,
                defaults={"color_field": color},
            )
            layer_table[layer.dxf.name] = {
                "layer_obj": layer_obj,
                "geometries": [],
            }
        return layer_table

    def extract_entities(self, msp, e_type, m, utm2world, layer_table):
        for e in msp.query(e_type):
            geo_proxy = get_geo_proxy(e, m, utm2world)
            if geo_proxy:
                if e_type in ["LWPOLYLINE", "POLYLINE"]:
                    entity_data = {}
                    # check if it's a true polygon
                    try:
                        poly = Polygon(e.vertices_in_wcs())
                        # look for texts in same layer
                        for t_type in self.text_types:
                            txts = msp.query(f"{t_type}[layer=='{e.dxf.layer}']")
                            for t in txts:
                                point = Point(t.dxf.insert)
                                # check if text is contained by polygon
                                if poly.contains(point):
                                    # handle different type of texts
                                    if t_type == "TEXT":
                                        entity_data["Name"] = t.dxf.text
                                    else:
                                        entity_data["Name"] = t.text
                                    break
                        if e.is_closed:
                            entity_data["Surface"] = round(poly.area, 2)
                        if e.dxf.thickness:
                            entity_data["Height"] = round(e.dxf.thickness, 2)
                        entity_data["Perimeter"] = round(poly.length, 2)
                        if e.dxf.const_width:
                            entity_data["Width"] = round(e.dxf.const_width, 2)
                        ent = Entity.objects.create(
                            layer=layer_table[e.dxf.layer]["layer_obj"],
                            geom={
                                "geometries": [geo_proxy.__geo_interface__],
                                "type": "GeometryCollection",
                            },
                            # data=entity_data,
                        )
                        for key, value in entity_data.items():
                            EntityData.objects.create(
                                entity=ent,
                                key=key,
                                value=value,
                            )
                    except (AttributeError, ValueError):
                        # not true polygon, add to layer entity
                        layer_table[e.dxf.layer]["geometries"].append(
                            geo_proxy.__geo_interface__
                        )
                else:
                    # not polyline, add to layer entity
                    layer_table[e.dxf.layer]["geometries"].append(
                        geo_proxy.__geo_interface__
                    )

    def create_layer_entities(self, layer_table):
        for name, layer_data in layer_table.items():
            # next conditional is true TDD!
            if len(layer_data["geometries"]) == 0:
                continue
            Entity.objects.create(
                layer=layer_data["layer_obj"],
                geom={
                    "geometries": layer_data["geometries"],
                    "type": "GeometryCollection",
                },
            )

    def save_blocks(self, doc, m, utm2world):
        block_table = {}
        for block in doc.blocks:
            if block.name in self.name_blacklist:
                continue
            geometries = []
            for e_type in self.entity_types:
                # extract entities
                for e in block.query(e_type):
                    geo_proxy = get_geo_proxy(e, m, utm2world)
                    if geo_proxy:
                        geometries.append(geo_proxy.__geo_interface__)
            # create block as Layer
            if not geometries == []:
                # use get or create to pass tests
                block_obj, created = Layer.objects.get_or_create(
                    drawing_id=self.id,
                    name=block.name,
                    is_block=True,
                    defaults={
                        "geom": {
                            "geometries": geometries,
                            "type": "GeometryCollection",
                        }
                    },
                )
                block_table[block.name] = block_obj
        return block_table

    def extract_insertions(self, ins, msp, m, utm2world, layer_table, block_table):
        # filter blacklisted blocks
        if ins.dxf.name in self.name_blacklist:
            return
        point = msp.add_point(ins.dxf.insert)
        geo_proxy = get_geo_proxy(point, m, utm2world)
        if geo_proxy:
            insertion_point = geo_proxy.__geo_interface__
        geometries = []
        # 'generator' object has no attribute 'query'
        for e in ins.virtual_entities():
            if e.dxftype() in self.entity_types:
                # extract entity
                geo_proxy = get_geo_proxy(e, m, utm2world)
                if geo_proxy:
                    geometries.append(geo_proxy.__geo_interface__)
        # prepare block data
        if ins.dxf.rotation:
            rotation = round(ins.dxf.rotation, 2)
        else:
            rotation = 0
        if ins.dxf.xscale:
            xscale = round(ins.dxf.xscale, 2)
        else:
            xscale = 1
        if ins.dxf.yscale:
            yscale = round(ins.dxf.yscale, 2)
        else:
            yscale = 1
        # create Insertion
        ins_obj = Entity.objects.create(
            layer=layer_table[ins.dxf.layer]["layer_obj"],
            block=block_table[ins.dxf.name],
            insertion=insertion_point,
            geom={
                "geometries": geometries,
                "type": "GeometryCollection",
            },
            rotation=rotation,
            xscale=xscale,
            yscale=yscale,
        )
        # add attributes
        if ins.attribs:
            for attr in ins.attribs:
                EntityData.objects.create(
                    entity=ins_obj,
                    key=attr.dxf.tag,
                    value=attr.dxf.text,
                )

    def write_csv(self, writer):
        writer_data = []
        layers = self.related_layers.all()
        for layer in layers:
            entities = layer.related_entities.all()
            for e in entities:
                if not e.related_data:
                    continue
                entity_data = {
                    "id": e.id,
                    "layer": layer.name,
                }
                if e.insertion:
                    entity_data["Latitude"] = e.insertion["coordinates"][0]
                    entity_data["Longitude"] = e.insertion["coordinates"][1]
                    entity_data["Block"] = e.block.name
                    entity_data["X scale"] = e.xscale
                    entity_data["Y scale"] = e.yscale
                    entity_data["Rotation"] = e.rotation
                    for ed in e.related_data.all():
                        entity_data["attributes"] = {}
                        entity_data["attributes"][ed.key] = ed.value
                else:
                    for ed in e.related_data.all():
                        entity_data[ed.key] = ed.value
                writer_data.append(entity_data)
        writer.writerow(
            [
                _("ID"),
                _("Layer"),
                _("Block"),
                _("Name"),
                _("Surface"),
                _("Perimeter"),
                _("Height"),
                _("Width"),
                _("Rotation"),
                _("X scale"),
                _("Y scale"),
                _("Latitude"),
                _("Longitude"),
                _("Attributes"),
            ]
        )
        keys = [
            "Name",
            "Surface",
            "Perimeter",
            "Height",
            "Width",
        ]
        for wd in writer_data:
            row = []
            row.append(wd["id"])
            row.append(wd["layer"])
            if "Block" in wd:
                row.append(wd["Block"])
            else:
                row.append("")
            for k in keys:
                if k in wd:
                    row.append(wd[k])
                else:
                    row.append("")
            if "Rotation" in wd:
                row.append(wd["Rotation"])
            else:
                row.append("")
            if "X scale" in wd:
                row.append(wd["X scale"])
            else:
                row.append("")
            if "Y scale" in wd:
                row.append(wd["Y scale"])
            else:
                row.append("")
            if "Latitude" in wd:
                row.append(wd["Latitude"])
            else:
                row.append("")
            if "Longitude" in wd:
                row.append(wd["Longitude"])
            else:
                row.append("")
            if "attributes" in wd:
                for key, value in wd["attributes"].items():
                    row.append(key)
                    row.append(value)
            writer.writerow(row)
        return writer

    def prepare_dxf_to_download(self):
        blocks = self.related_layers.filter(is_block=True)
        if not blocks.exists():
            return
        block_list = blocks.values_list("id", flat=True)
        # extract entities to be processed
        entities = Entity.objects.filter(
            block_id__in=block_list, data__added=True
        ).prefetch_related()
        if not entities.exists():
            return
        # prepare transformers
        world2utm, utm2world, utm_wcs, rot = self.prepare_transformers()
        # start DXF
        doc = ezdxf.readfile(self.dxf.path)
        msp = doc.modelspace()
        geodata = msp.get_geodata()
        # get transform matrix from geodata
        m, epsg = geodata.get_crs_transformation(no_checks=True)
        # add insertions
        for ent in entities:
            # could be a new layer
            if ent.layer.name not in doc.layers:
                new_layer = doc.layers.add(ent.layer.name)
                color = ImageColor.getcolor(ent.layer.color_field, "RGB")
                new_layer.rgb = color
            geo_proxy = geo.GeoProxy.parse(ent.insertion)
            geo_proxy.apply(lambda v: ezdxf.math.Vec3(world2utm.transform(v.x, v.y)))
            geo_proxy.crs_to_wcs(m)
            for entity in geo_proxy.to_dxf_entities():
                point = entity.dxf.location
            block_ref = msp.add_blockref(
                ent.block.name,
                point,
                dxfattribs={
                    "xscale": ent.xscale,
                    "yscale": ent.yscale,
                    "rotation": ent.rotation,
                    "layer": ent.layer.name,
                },
            )
            # add block attributes
            values = {}
            for ed in ent.related_data.all():
                values[ed.key] = ed.value
            block_ref.add_auto_attribs(values)
            # change JSON so entity is not selected again
            ent.data["added"] = False
        # update all entities
        Entity.objects.bulk_update(entities, ["data"])
        # replace dxf
        doc.saveas(filename=self.dxf.path, encoding="utf-8", fmt="asc")


class Layer(models.Model):

    drawing = models.ForeignKey(
        Drawing,
        on_delete=models.CASCADE,
        related_name="related_layers",
        verbose_name=_("Drawing"),
    )
    name = models.CharField(
        _("Layer name"),
        max_length=50,
    )
    color_field = ColorField(default="#FFFFFF")
    linetype = models.BooleanField(
        _("Continuous linetype"),
        default=True,
    )
    is_block = models.BooleanField(
        default=False,
        editable=False,
    )
    geom = GeometryCollectionField(
        null=True,
    )

    class Meta:
        verbose_name = _("Layer")
        verbose_name_plural = _("Layers")
        ordering = ("name",)
        constraints = [
            models.UniqueConstraint(
                fields=["drawing", "name", "is_block"], name="unique_layer_name"
            ),
        ]

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # check for layer unique name
        try:
            # avoid TransactionManagementError
            with transaction.atomic():
                super().save(*args, **kwargs)
        except IntegrityError:
            self.name = f"{self.name}_{get_random_string(7)}"
            super().save(*args, **kwargs)


def get_default_entity_data():
    return {"processed": "true"}


class Entity(models.Model):

    layer = models.ForeignKey(
        Layer,
        on_delete=models.CASCADE,
        related_name="related_entities",
    )
    data = models.JSONField(
        default=get_default_entity_data,
    )
    geom = GeometryCollectionField(
        null=True,
    )
    block = models.ForeignKey(
        Layer,
        on_delete=models.CASCADE,
        related_name="related_insertions",
        null=True,
        blank=False,
    )
    insertion = PointField(
        null=True,
    )
    xscale = models.FloatField(
        _("X scale"),
        default=1,
    )
    yscale = models.FloatField(
        _("Y scale"),
        default=1,
    )
    rotation = models.FloatField(
        _("Rotation"),
        default=0,
    )

    class Meta:
        verbose_name = _("Entity")
        verbose_name_plural = _("Entities")

    @property
    def popupContent(self):
        if self.block:
            url = reverse("djeocad:insertion_change", kwargs={"pk": self.id})
            title_str = f'<p><a href="{url}">ID = {self.id}</a></p>'
        else:
            title_str = f"<p>ID = {self.id}</p>"
        ltype = _("Layer")
        title_str += f"<ul><li>{ltype}: {nh3.clean(self.layer.name)}</li>"
        if self.block:
            ltype = _("Block")
            title_str += f"<li>{ltype}: {nh3.clean(self.block.name)}</li>"
        data = ""
        ent_data = self.related_data.all()
        if ent_data.exists():
            if self.block:
                data += "</ul><p>Attributes</p><ul>"
                for ed in ent_data:
                    data += f"<li>{nh3.clean(ed.key)} = {nh3.clean(ed.value)}</li>"
            else:
                for ed in ent_data:
                    data += f"<li>{nh3.clean(ed.key)} = {nh3.clean(ed.value)}</li>"
        data += "</ul>"
        return {
            "content": title_str + data,
            "color": self.layer.color_field,
            "linetype": self.layer.linetype,
            "layer": _("Layer - ") + nh3.clean(self.layer.name),
        }

    def save(self, *args, **kwargs):
        if "added" in self.data and self.block:
            # we will use a fake DXF to help us
            # prepare transformers
            world2utm, utm2world, utm_wcs, rot = (
                self.block.drawing.prepare_transformers()
            )
            # start fake DXF
            doc = ezdxf.new()
            msp = doc.modelspace()
            # we fake geodata
            geodata = msp.new_geodata()
            geodata = self.block.drawing.fake_geodata(geodata, utm_wcs, rot)
            # get transform matrix from fake geodata
            m, epsg = geodata.get_crs_transformation(no_checks=True)
            # add block to fake DXF
            block = doc.blocks.new(name=self.block.name)
            geometries = self.block.geom["geometries"]
            for geom in geometries:
                geo_proxy = geo.GeoProxy.parse(geom)
                geo_proxy.apply(
                    lambda v: ezdxf.math.Vec3(world2utm.transform(v.x, v.y))
                )
                geo_proxy.crs_to_wcs(m)
                for entity in geo_proxy.to_dxf_entities(dxfattribs={"layer": "0"}):
                    block.add_entity(entity)
            geo_proxy = geo.GeoProxy.parse(self.insertion)
            geo_proxy.apply(lambda v: ezdxf.math.Vec3(world2utm.transform(v.x, v.y)))
            geo_proxy.crs_to_wcs(m)
            for entity in geo_proxy.to_dxf_entities():
                point = entity.dxf.location
            instance = msp.add_blockref(
                self.block.name,
                point,
                dxfattribs={
                    "xscale": self.xscale,
                    "yscale": self.yscale,
                    "rotation": self.rotation,
                    "layer": self.layer.name,
                },
            )
            # use fake instance to generate new geometries
            geometries = []
            # 'generator' object has no attribute 'query'
            for e in instance.virtual_entities():
                if e.dxftype() in self.block.drawing.entity_types:
                    # extract entity
                    geo_proxy = get_geo_proxy(e, m, utm2world)
                    if geo_proxy:
                        geometries.append(geo_proxy.__geo_interface__)
            # update Insertion
            self.geom = {
                "geometries": geometries,
                "type": "GeometryCollection",
            }
        super().save(*args, **kwargs)
        if self.block and not self.related_data.exists():
            first = Entity.objects.filter(block=self.block).first()
            data = first.related_data.all()
            if data.exists():
                for d in data:
                    EntityData.objects.create(
                        entity=self,
                        key=d.key,
                        value=d.value,
                    )


class EntityData(models.Model):

    entity = models.ForeignKey(
        Entity,
        on_delete=models.CASCADE,
        related_name="related_data",
    )
    key = models.CharField(
        _("Data key"),
        max_length=50,
    )
    value = models.CharField(
        _("Data value"),
        max_length=100,
    )

    class Meta:
        verbose_name = _("Entity Data")
        verbose_name_plural = _("Entity Data")


"""
    Collection of utilities
"""


def cad2hex(color):
    if isinstance(color, tuple):
        return "#{:02x}{:02x}{:02x}".format(color[0], color[1], color[2])
    rgb24 = ezdxf.colors.DXF_DEFAULT_COLORS[color]
    return "#{:06X}".format(rgb24)


def get_geo_proxy(entity, matrix, transformer):
    geo_proxy = geo.proxy(entity)
    if geo_proxy.geotype == "Polygon":
        if not shape(geo_proxy).is_valid:
            return False
    geo_proxy.wcs_to_crs(matrix)
    geo_proxy.apply(lambda v: ezdxf.math.Vec3(transformer.transform(v.x, v.y)))
    return geo_proxy



from django.apps import AppConfig
from django.db.models.signals import post_migrate
from django.utils.translation import gettext as _


def create_djeocad_group(sender, **kwargs):
    from django.contrib.auth.models import Group, Permission
    from django.contrib.contenttypes.models import ContentType

    grp, created = Group.objects.get_or_create(name=_("GeoCAD Manager"))
    if created:
        types = ContentType.objects.filter(app_label="djeocad").values_list(
            "id", flat=True
        )
        permissions = Permission.objects.filter(content_type_id__in=types)
        grp.permissions.set(permissions)


class DjeocadConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "djeocad"

    def ready(self):
        post_migrate.connect(create_djeocad_group, sender=self)



# Generated by Django 5.1.1 on 2024-09-17 17:17

import django.core.validators
import django.db.models.deletion
import djgeojson.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="drawing",
            options={"verbose_name": "Drawing", "verbose_name_plural": "Drawings"},
        ),
        migrations.AddField(
            model_name="drawing",
            name="designx",
            field=models.FloatField(default=0, verbose_name="Design point X..."),
        ),
        migrations.AddField(
            model_name="drawing",
            name="designy",
            field=models.FloatField(default=0, verbose_name="...Y"),
        ),
        migrations.AddField(
            model_name="drawing",
            name="dxf",
            field=models.FileField(
                default="uploads/djeocad/dxf/sample.dxf",
                max_length=200,
                upload_to="uploads/djeocad/dxf/",
                validators=[
                    django.core.validators.FileExtensionValidator(
                        allowed_extensions=["dxf"]
                    )
                ],
                verbose_name="DXF file",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="drawing",
            name="epsg",
            field=models.IntegerField(
                editable=False, null=True, verbose_name="CRS code"
            ),
        ),
        migrations.AddField(
            model_name="drawing",
            name="geom",
            field=djgeojson.fields.PointField(
                blank=True, null=True, verbose_name="Location"
            ),
        ),
        migrations.AddField(
            model_name="drawing",
            name="image",
            field=models.ImageField(
                blank=True,
                null=True,
                upload_to="uploads/djeocad/images/",
                verbose_name="Image",
            ),
        ),
        migrations.AddField(
            model_name="drawing",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="parent_drawing",
                to="djeocad.drawing",
                verbose_name="Parent Drawing",
            ),
        ),
        migrations.AddField(
            model_name="drawing",
            name="rotation",
            field=models.FloatField(default=0, verbose_name="Rotation"),
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-14 17:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0005_entity_block"),
    ]

    operations = [
        migrations.CreateModel(
            name="EntityData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=50, verbose_name="Data key")),
                ("value", models.CharField(max_length=100, verbose_name="Data value")),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="related_data",
                        to="djeocad.entity",
                    ),
                ),
            ],
            options={
                "verbose_name": "Entity Data",
                "verbose_name_plural": "Entity Data",
            },
        ),
    ]



import csv
from typing import Any

from django.contrib.auth.decorators import permission_required
from django.db.models.query import QuerySet
from django.forms import FloatField, ModelForm, NumberInput
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, ListView

from .models import Drawing, Entity, EntityData


class DrawingListView(ListView):
    model = Drawing
    template_name = "djeocad/drawing_list.html"

    def get_queryset(self) -> QuerySet[Any]:
        qs = Drawing.objects.exclude(epsg=None)
        return qs

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["unreferenced"] = Drawing.objects.filter(epsg=None)
        return context


class DrawingDetailView(DetailView):
    model = Drawing
    template_name = "djeocad/drawing_detail.html"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        layers = self.object.related_layers.filter(is_block=False)
        if self.object.related_layers.filter(is_block=True).exists():
            context["blocks"] = True
        id_list = layers.values_list("id", flat=True)
        context["lines"] = Entity.objects.filter(
            layer_id__in=id_list
        ).prefetch_related()
        name_list = layers.values_list("name", flat=True)
        context["layer_list"] = list(dict.fromkeys(name_list))
        context["layer_list"] = [_("Layer - ") + s for s in context["layer_list"]]
        return context


class EntityCreateForm(ModelForm):
    lat = FloatField(
        label=_("Latitude"),
        widget=NumberInput(attrs={"max": 90, "min": -90}),
        required=True,
    )
    long = FloatField(
        label=_("Longitude"),
        widget=NumberInput(attrs={"max": 180, "min": -180}),
        required=True,
    )

    class Meta:
        model = Entity
        fields = ["layer", "block", "rotation", "xscale", "yscale"]

    def clean(self):
        cleaned_data = super().clean()
        lat = cleaned_data["lat"]
        long = cleaned_data["long"]
        if lat > 90 or lat < -90:
            self.add_error("lat", _("Invalid value"))
        if long > 180 or long < -180:
            self.add_error("long", _("Invalid value"))
        return cleaned_data


@permission_required("djeocad.change_drawing")
def add_block_insertion(request, pk):
    drawing = get_object_or_404(Drawing, id=pk)
    blocks = drawing.related_layers.filter(is_block=True)
    if not blocks.exists():
        raise Http404
    layers = drawing.related_layers.filter(is_block=False)
    context = {}
    if request.POST:
        form = EntityCreateForm(request.POST)
        if form.is_valid():
            ent = Entity(
                layer=form.cleaned_data["layer"],
                block=form.cleaned_data["block"],
                rotation=form.cleaned_data["rotation"],
                xscale=form.cleaned_data["xscale"],
                yscale=form.cleaned_data["yscale"],
                insertion={
                    "type": "Point",
                    "coordinates": [
                        form.cleaned_data["long"],
                        form.cleaned_data["lat"],
                    ],
                },
                data={
                    "processed": "true",
                    "added": "true",
                },
            )
            ent.save()
            return HttpResponseRedirect(
                reverse("djeocad:insertion_change", kwargs={"pk": ent.id})
            )
    else:
        form = EntityCreateForm(
            initial={
                "rotation": 0,
                "xscale": 1,
                "yscale": 1,
                "lat": drawing.geom["coordinates"][1],
                "long": drawing.geom["coordinates"][0],
            }
        )
    form.fields["layer"].queryset = layers
    form.fields["block"].queryset = blocks
    context["form"] = form
    id_list = layers.values_list("id", flat=True)
    context["lines"] = Entity.objects.filter(layer_id__in=id_list).prefetch_related()
    name_list = layers.values_list("name", flat=True)
    context["layer_list"] = list(dict.fromkeys(name_list))
    context["layer_list"] = [_("Layer - ") + s for s in context["layer_list"]]
    context["drawing"] = drawing
    return TemplateResponse(request, "djeocad/entity_create.html", context)


@permission_required("djeocad.change_drawing")
def change_block_insertion(request, pk):
    object = get_object_or_404(Entity, id=pk)
    drawing = object.layer.drawing
    blocks = drawing.related_layers.filter(is_block=True)
    if not blocks.exists():
        raise Http404
    layers = drawing.related_layers.filter(is_block=False)
    context = {}
    if request.POST:
        form = EntityCreateForm(request.POST)
        if form.is_valid():
            object.layer = form.cleaned_data["layer"]
            object.block = form.cleaned_data["block"]
            object.rotation = form.cleaned_data["rotation"]
            object.xscale = form.cleaned_data["xscale"]
            object.yscale = form.cleaned_data["yscale"]
            object.insertion = {
                "type": "Point",
                "coordinates": [
                    form.cleaned_data["long"],
                    form.cleaned_data["lat"],
                ],
            }
            object.save()
            return HttpResponseRedirect(
                reverse("djeocad:drawing_detail", kwargs={"pk": drawing.id})
            )
    else:
        form = EntityCreateForm(
            initial={
                "layer": object.layer,
                "block": object.block,
                "rotation": object.rotation,
                "xscale": object.xscale,
                "yscale": object.yscale,
                "lat": object.insertion["coordinates"][1],
                "long": object.insertion["coordinates"][0],
            }
        )
    form.fields["layer"].queryset = layers
    form.fields["block"].queryset = blocks
    context["form"] = form
    id_list = layers.values_list("id", flat=True)
    context["lines"] = Entity.objects.filter(layer_id__in=id_list).prefetch_related()
    name_list = layers.values_list("name", flat=True)
    context["layer_list"] = list(dict.fromkeys(name_list))
    context["layer_list"] = [_("Layer - ") + s for s in context["layer_list"]]
    context["drawing"] = drawing
    context["object"] = object
    context["related_data"] = object.related_data.all()
    context["data_form"] = EntityDataForm()
    return TemplateResponse(request, "djeocad/entity_change.html", context)


@permission_required("djeocad.change_drawing")
def delete_block_insertion(request, pk):
    object = get_object_or_404(Entity, id=pk)
    drawing = object.layer.drawing
    object.delete()
    return HttpResponseRedirect(
        reverse("djeocad:drawing_detail", kwargs={"pk": drawing.id})
    )


class EntityDataListView(ListView):
    model = EntityData
    template_name = "djeocad/htmx/entity_data_list.html"
    context_object_name = "related_data"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.entity = get_object_or_404(Entity, id=kwargs["pk"])

    def get_queryset(self):
        return EntityData.objects.filter(entity=self.entity)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["object"] = self.entity
        context["data_form"] = EntityDataForm()
        return context


class EntityDataForm(ModelForm):
    class Meta:
        model = EntityData
        fields = ["key", "value"]


@permission_required("djeocad.change_drawing")
def create_entity_data(request, pk):
    if (
        "Hx-Request" not in request.headers
        or not request.headers["Hx-Request"] == "true"
    ):
        raise Http404
    entity = get_object_or_404(Entity, id=pk)
    if request.POST:
        form = EntityDataForm(request.POST)
        if form.is_valid():
            EntityData.objects.create(
                entity=entity,
                key=form.cleaned_data["key"],
                value=form.cleaned_data["value"],
            )
            return HttpResponseRedirect(
                reverse("djeocad:data_list", kwargs={"pk": entity.id})
            )


@permission_required("djeocad.change_drawing")
def delete_entity_data(request, pk):
    if (
        "Hx-Request" not in request.headers
        or not request.headers["Hx-Request"] == "true"
    ):
        raise Http404
    object = get_object_or_404(EntityData, id=pk)
    entity = object.entity
    object.delete()
    return HttpResponseRedirect(reverse("djeocad:data_list", kwargs={"pk": entity.id}))


def csv_download(request, pk):
    drawing = get_object_or_404(Drawing, id=pk)
    # Create the HttpResponse object with the appropriate CSV header.
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = f'attachment; filename="{drawing.title}.csv"'
    writer = csv.writer(response)
    writer = drawing.write_csv(writer)

    return response


def drawing_download(request, pk):
    drawing = get_object_or_404(Drawing, id=pk)
    drawing.prepare_dxf_to_download()
    response = HttpResponse(drawing.dxf, content_type="text/plain")
    response["Content-Disposition"] = f"attachment; filename={drawing.title}.dxf"

    return response



# Generated by Django 5.1.1 on 2024-11-14 18:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0006_entitydata"),
    ]

    operations = [
        migrations.AddField(
            model_name="entity",
            name="rotation",
            field=models.FloatField(default=0, verbose_name="Rotation"),
        ),
        migrations.AddField(
            model_name="entity",
            name="xscale",
            field=models.FloatField(default=1, verbose_name="X scale"),
        ),
        migrations.AddField(
            model_name="entity",
            name="yscale",
            field=models.FloatField(default=1, verbose_name="Y scale"),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-18 13:45

import django.db.models.deletion
import djgeojson.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0003_layer"),
    ]

    operations = [
        migrations.CreateModel(
            name="Entity",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("data", models.JSONField(null=True)),
                ("geom", djgeojson.fields.GeometryCollectionField()),
                ("insertion", djgeojson.fields.PointField(null=True)),
                (
                    "layer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="related_entities",
                        to="djeocad.layer",
                    ),
                ),
            ],
            options={
                "verbose_name": "Entity",
                "verbose_name_plural": "Entities",
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-19 18:52

import djgeojson.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0008_alter_entity_data"),
    ]

    operations = [
        migrations.AlterField(
            model_name="entity",
            name="geom",
            field=djgeojson.fields.GeometryCollectionField(null=True),
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-01 15:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djeocad", "0009_alter_entity_geom"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="layer",
            constraint=models.UniqueConstraint(
                fields=("drawing", "name", "is_block"), name="unique_layer_name"
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-16 18:57

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Drawing",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="Name of the drawing",
                        max_length=50,
                        verbose_name="Name",
                    ),
                ),
            ],
        ),
    ]



from django.urls import path

from .views import (
    DrawingDetailView,
    DrawingListView,
    EntityDataListView,
    add_block_insertion,
    change_block_insertion,
    create_entity_data,
    csv_download,
    delete_block_insertion,
    delete_entity_data,
    drawing_download,
)

app_name = "djeocad"
urlpatterns = [
    path("", DrawingListView.as_view(), name="drawing_list"),
    path("<pk>", DrawingDetailView.as_view(), name="drawing_detail"),
    path("<pk>/insertion", add_block_insertion, name="insertion_create"),
    path("insertion/<pk>/change", change_block_insertion, name="insertion_change"),
    path("insertion/<pk>/delete", delete_block_insertion, name="insertion_delete"),
    path("insertion/<pk>/data-list", EntityDataListView.as_view(), name="data_list"),
    path("insertion/<pk>/data-create", create_entity_data, name="data_create"),
    path("entity-data/<pk>/delete", delete_entity_data, name="data_delete"),
    path(
        "<pk>/csv",
        csv_download,
        name="drawing_csv",
    ),
    path(
        "<pk>/download",
        drawing_download,
        name="drawing_download",
    ),
]
